# Copyright 2016 Jochen Fabricius
# Distribution and modification is encouraged

package provide cmatrix 1.0

#lappend auto_path [file join [glob ~] Projects FR lib]
package require complex

namespace eval ::cmatrix {
}

proc ::cmatrix::toStr {M} {
  set r [llength ${M}]
  set c [llength [lindex ${M} 0]]
  for {set i 0} {$i < $r} {incr i} {
    puts [lindex $M $i]
  }
}

proc ::cmatrix::create {M rows cols} {
  upvar $M _M
  set _M [list]
  for {set i 0} {$i < $rows} {incr i} {
    set c [list]
    for {set j 0} {$j < $cols} {incr j} {
      lappend c [::complex::cset C]
    }
    lappend _M $c
  }

  return ${_M}
}

proc ::cmatrix::mset {M row col val} {
  upvar $M _M
  set r [llength ${_M}]
  set c [llength [lindex ${_M} 0]]
  if {($row < 0 && $col < 0) || $row > $r || $col > $c} {
    return -code error {bad index}
  }
  if {$col < 0} {
    # replace row
    if {[llength $val] > $c} {
      return -code error {bad size}
    }
    # make sure that every element is complex
    set j 0
    foreach {Z} $val {
	    if {$j >= $c} {break}
	    lassign $Z re im
	    if {$re eq {}} {set re 0.0}
	    if {$im eq {}} {set im 0.0}
	    lset _M $row $j [::complex::cset Z $re $im]
	    incr j
    }
    for {} {$j < $c} {incr j} {
      lset _M $row $j [::complex::cset Z]
    }
  } elseif {$row < 0} {
    # replace column
    if {[llength $val] > $r} {
      return -code error {bad size}
    }
    # make sure that every element is complex
    set i 0
    foreach {Z} $val {
	    if {$i >= $r} {break}
	    lassign $Z re im
	    if {$re eq {}} {set re 0.0}
	    if {$im eq {}} {set im 0.0}
	    lset _M $i $col [::complex::cset Z $re $im]
	    incr i
    }
    for {} {$i < $r} {incr i} {
      lset _M $i $col [::complex::cset Z]
    }
  } else {
    lassign $val re im
    lset _M $row $col [::complex::cset Z $re $im]
  }
  return ${_M}
}

proc ::cmatrix::madd {A B {anchor {}}} {
	upvar $A _A
	set ma [llength ${_A}]
	set na [llength [lindex ${_A} 0]]
	set mb [llength $B]
	set nb [llength [lindex $B 0]]
	if {$anchor eq {}} {
		# standard matrix addition, both matrices must be of same type
		if {$ma != $mb || $na != $nb} {
			return -code error {type mismatch}
		}
		for {set i 0} {$i < $ma} {incr i} {
			lset _A $i [::complex::cvadd [lindex ${_A} $i] [lindex $B $i]]
		}
	} else {
		if {$ma < $mb || $na < $nb} {
			return -code error {type mismatch}
		}
		switch -- $anchor {
			lt {
				set mo [set no 0]
			}
			lb {
				set mo [expr {$ma-$mb}]
				set no 0
			}
			rt {
				set mo 0
				set no [expr {$na-$nb}]
			}
			rb {
				set mo [expr {$ma-$mb}]
				set no [expr {$na-$nb}]
			}
		}
		for {set i $mo; set ib 0} {$ib < $mb} {incr i; incr ib} {
			for {set j $no; set jb 0} {$jb < $nb} {incr j; incr jb} {
				lset _A $i $j [::complex::cadd [lindex ${_A} $i $j] [lindex $B $ib $jb]]
			}
		}
	}

}

proc ::cmatrix::mmul {A B} {
	set ma [llength ${A}]
	set na [llength [lindex ${A} 0]]
	set mb [llength $B]
	set nb [llength [lindex $B 0]]
	if {$mb == 2 && [llength [lindex $B 0 0]] == 1} {
		# B is a scalar, use simple scalar multiplication
		::cmatrix::create M $ma $na
		for {set i 0} {$i < $ma} {incr i} {
			for {set j 0} {$j < $na} {incr j} {
				lset M $i $j [::complex::cmul [lindex $A $i $j] $B]
			}
		}
	} else {
		if {$na != $mb} {
			return -code error {size mismatch}
		}
		::cmatrix::create M $ma $nb
		# create column vectors (same as transposed B)
		set CV [list]
		for {set j 0} {$j < $nb} {incr j} {
			set cv [list]
			for {set k 0} {$k < $mb} {incr k} {
				lappend cv [lindex $B $k $j]
			}
			lappend CV $cv
		}

		for {set i 0} {$i < $ma} {incr i} {
			for {set j 0} {$j < $nb} {incr j} {
				# vector multiplication
				lset M $i $j [::complex::cvsum [::complex::cvmul [lindex $A $i] [lindex $CV $j]]] 
			}
		}
	}
	return $M
}

proc ::cmatrix::transpose {A} {
	set ma [llength ${A}]
	set na [llength [lindex ${A} 0]]
	set M [list]
	for {set j 0} {$j < $na} {incr j} {
		set cv [list]
		for {set k 0} {$k < $ma} {incr k} {
			lappend cv [lindex $A $k $j]
		}
		lappend M $cv
	}
	return $M
}

# returns a permutation matrix and index vector which pivotises A
proc ::cmatrix::pivot {A} {
	set ma [llength ${A}]
	set na [llength [lindex ${A} 0]]
	for {set j 0} {$j < $ma && $j < $na} {incr j} {
		lappend p $j
	}
	for {set j 0} {$j < $ma-1 && $j < $na-1} {incr j} {
		# find largest element in column
		set max 0.0
		set maxI -1
		for {set i $j} {$i < $ma} {incr i} {
			if {[set abs [::complex::cabs [lindex $A [lindex $p $i] $j]]] > $max} {
				set max $abs
				set maxI $i
			}
		}
		if {$maxI >= 0} {
			set t [lindex $p $j]
			lset p $j [lindex $p $maxI]
			lset p $maxI $t
		}
	}
	# set up permutation matrix
	set m [llength $p]
	::cmatrix::create P $m $m
	for {set i 0} {$i < $m} {incr i} {
		lset P $i [lindex $p $i] {1.0 0.0}
	}
	return [list $P $p]
}

# performs a LU decomposition
# stores U in A and returns L
proc ::cmatrix::LU {A} {
	upvar $A _A
	set ma [llength ${_A}]
	set na [llength [lindex ${_A} 0]]
	# create the null matrix L; note the switched number of rows and columns so A=L*R holds true
	::cmatrix::create L $na $ma
	for {set j 0} {$j < $ma && $j < $na} {incr j} {
		lset L $j $j {1.0 0.0}
		if {[::complex::cabs [set a [lindex ${_A} $j $j]]] != 0} {
			for {set i [expr {$j+1}]} {$i < $ma} {incr i} {
				set rv [lindex ${_A} $i]
				set k [::complex::cdiv [lindex $rv $j] $a]
				lset L $i $j $k
				lset _A $i [::complex::cvsub $rv [::complex::cvmul [lindex ${_A} $j] $k]]
			}
		}
	}
	return $L
}

proc ::cmatrix::solve {A _x b} {
	upvar ${_x} x
	set ma [llength $A]
	set na [llength [lindex $A 0]]

	# pivotise matrix and result vector
	lassign [::cmatrix::pivot $A] P p
	set A [::cmatrix::mmul $P $A]
	set b [::cmatrix::mmul $P $b]

	# create the augmented matrix
	for {set i 0} {$i < $ma} {incr i} {
		set rv [lindex $A $i]
		lappend rv [lindex $b $i 0]
		lset A $i $rv
	}

	set L [::cmatrix::LU A]
	
	# solving by back substitution
	for {set i [expr {$ma-1}]; set j [expr {$na-1}]} {$i > 0 && $j > 0} {incr j -1} {
		# solve for x(i)
		lset A $i $na [set xi [::complex::cdiv [lindex $A $i $na] [lindex $A $i $j]]]
		# substitute x(i) in x(i-1)
		incr i -1
		lset A $i $j [set xi [::complex::cmul $xi [lindex $A $i $j]]]
		lset A $i $na [::complex::csub [lindex $A $i $na] $xi]
	}	
	# solve for x(0)
	lset A $i $na [set xi [::complex::cdiv [lindex $A $i $na] [lindex $A $i $j]]]
	for {set i 0} {$i < $ma} {incr i} {
		lset x $i 0 [lindex $A [lindex $p $i] $na]
	}
	return [list $P $p]
}

proc ::cmatrix::lsq {X b y} {

} 

proc test {} {
	::cmatrix::create A 2 2
	::cmatrix::mset A 0 -1 [list [::complex::cset Z 3] [::complex::cset Z 3]]
	::cmatrix::mset A 1 -1 [list [::complex::cset Z 1] [::complex::cset Z 2]]

	::cmatrix::create b 2 1
	::cmatrix::mset b -1 0 [list [::complex::cset Z 5] [::complex::cset Z 4]]

	::cmatrix::create x 2 1
	::cmatrix::solve $A x $b
	puts "x="
	::cmatrix::toStr $x
	puts "A*x="
	::cmatrix::toStr [::cmatrix::mmul $A $x]

}

#test
